/*
* Copyright c                  Realtek Semiconductor Corporation, 2002  
* All rights reserved.
* 
* Program : 
* Abstract : 
* Creator : 
* Author :  
* $Id: rtl8651_alg_nbt.c,v 1.1.1.1 2006/01/16 02:39:43 jason Exp $
* $Log: rtl8651_alg_nbt.c,v $
* Revision 1.1.1.1  2006/01/16 02:39:43  jason
* backup source,because now we directly use romedriver source code, i open a new directory;
*
* Revision 1.1.1.1  2005/11/21 05:20:03  chenyl
* +: WPNT834
* 	- System provider	: Sercomm
* 	- Brand Name		: Netgear
* 	- 3rt Party		: Airgo Gen 3
* 	- SDK			: 066-3
* 	- ROMEDRV		: 3.6-3
*
*
* Revision 1.4  2004/09/07 14:52:13  chenyl
* *: bug fix: napt-redirect fragment packet checksum-recalculate
* *: bug fix: conflict between protocol stack flow cache and protocol stack action
* *: bug fix: protocol stack action mis-process to UDP-zero-checksum packet
* *: separate the header file:
*         - internal : rtl8651_tblDrvFwdLocal.h
*         - external : rtl8651_tblDrvFwd.h
*
* Revision 1.3  2004/07/23 13:19:57  tony
* *: remove all warning messages.
*
* Revision 1.2  2004/04/20 03:44:03  tony
* if disable define "RTL865X_OVER_KERNEL" and "RTL865X_OVER_LINUX", __KERNEL__ and __linux__ will be undefined.
*
* Revision 1.1  2004/02/25 14:26:33  chhuang
* *** empty log message ***
*
* Revision 1.2  2004/02/25 14:24:52  chhuang
* *** empty log message ***
*
* Revision 1.5  2004/02/18 13:54:37  chhuang
* *** empty log message ***
*
* Revision 1.4  2003/12/19 11:40:01  tony
* change memcpy to rtl8651_memcpy
*
* Revision 1.3  2003/12/17 10:33:12  hiwu
* add checksum function
*
* Revision 1.2  2003/10/17 08:45:51  hiwu
* merge new function type
*
* Revision 1.1  2003/10/02 07:08:11  hiwu
* initial version
*
*/


#include "rtl_types.h"
#include "assert.h"
#include "types.h"
#include "rtl_queue.h"
#include "rtl_errno.h"
#include "mbuf.h"

#include "rtl8651_tblDrvFwdLocal.h"
#include "rtl8651_tblDrvProto.h"
#include "rtl8651_tblDrvLocal.h"
#include "rtl8651_alg_init.h"
#include "rtl8651_alg_nbt.h"


typedef struct
{
	struct in_addr oldaddr;
	uint16    oldport;
	struct in_addr newaddr;
	uint16    newport;
	uint16   *uh_sum;
	struct ip *pip;
} NBTArguments;

typedef struct
{
	uint8     type;
	uint8     flags;
	uint16    id;
	struct in_addr source_ip;
	uint16    source_port;
	uint16    len;
	uint16    offset;
} NbtDataHeader;

#define OpQuery		0
#define OpUnknown	4
#define OpRegist	5
#define OpRelease	6
#define OpWACK		7
#define OpRefresh	8
typedef struct
{
	uint16    nametrid;
	uint16    dir:1, opcode:4, nmflags:7, rcode:4;
	uint16    qdcount;
	uint16    ancount;
	uint16    nscount;
	uint16    arcount;
}
NbtNSHeader;

#define FMT_ERR		0x1
#define SRV_ERR		0x2
#define IMP_ERR		0x4
#define RFS_ERR		0x5
#define ACT_ERR		0x6
#define CFT_ERR		0x7

/* 
 * NetBios Datagram Handler (IP/UDP)
 */
#define DGM_DIRECT_UNIQ		0x10
#define DGM_DIRECT_GROUP	0x11
#define DGM_BROADCAST		0x12
#define DGM_ERROR			0x13
#define DGM_QUERY			0x14
#define DGM_POSITIVE_RES	0x15
#define DGM_NEGATIVE_RES	0x16


/* Question Section */
#define QS_TYPE_NB		0x0020
#define QS_TYPE_NBSTAT	0x0021
#define QS_CLAS_IN		0x0001
typedef struct
{
	uint16    type;					   /* The type of Request */
	uint16    class;				   /* The class of Request */
}
NBTNsQuestion;


/* Resource Record */
#define RR_TYPE_A		0x0001
#define RR_TYPE_NS		0x0002
#define RR_TYPE_NULL	0x000a
#define RR_TYPE_NB		0x0020
#define RR_TYPE_NBSTAT	0x0021
#define RR_CLAS_IN		0x0001
#define SizeOfNsResource	8
typedef struct
{
	uint16    type;
	uint16    class;
	uint32    ttl;
	uint16    rdlen;
}
NBTNsResource;

#define SizeOfNsRNB			6
typedef struct
{
	uint16    g:1, ont:2, resv:13;
	struct in_addr addr;
}
NBTNsRNB;

#define SizeOfResourceA		6
typedef struct
{
	struct in_addr addr;
}
NBTNsResourceA;

typedef struct
{
	uint16    opcode:4, flags:8, resv:4;
}
NBTNsResourceNULL;

typedef struct
{
	uint16    numnames;
}
NBTNsResourceNBSTAT;


#define MAX_NBT_NAME_LEN 16 

static uint8 *_rtl8651_l4AliasHandleName(uint8 * p, int8 * pmax, int8 *nbtname, int32 nbtlength);
static uint8 *_rtl8651_l4AliasHandleQuestion(uint16 , NBTNsQuestion *, int8 * , NBTArguments * );
static uint8 *_rtl8651_l4AliasHandleResource(uint16 , NBTNsResource * , int8 * , NBTArguments * );
static int32 _rtl8651_l4AliasHandleUdpNbt(struct ip *,struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *,struct in_addr *, uint16 );
static int32 _rtl8651_l4AliasHandleUdpNbtNS(struct ip * ,struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s * , struct in_addr * , uint16 * , struct in_addr * , uint16 * );


/*********************************************************************************
* Routine Name :  rtl8651_l4AliasHandleClientUdpNbtOut                                  
* Description NetBios Datagram Handler (IP/UDP):                                                 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
int32 rtl8651_l4AliasHandleClientUdpNbtOut(struct rtl_pktHdr *pkthdrPtr, struct ip *pip, struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{

	uint16    alias_port;
	struct in_addr alias_address;

	alias_address.s_addr = link->insideGlobalIpAddr;
	alias_port = link->insideGlobalPort;

	_rtl8651_l4AliasHandleUdpNbt(pip, link, &alias_address, alias_port);
	return SUCCESS;
}


/*********************************************************************************
* Routine Name :   rtl8651_l4AliasHandleUdpNbtNSOut                                
* Description :                                                 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
int32 rtl8651_l4AliasHandleClientUdpNbtNSOut(struct rtl_pktHdr *pkthdrPtr, struct ip *pip, struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)
{
	struct udphdr *ud;
	uint16    alias_port;
	struct in_addr alias_address;

	ud = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl & 0xf) << 2));

	alias_address.s_addr = link->insideGlobalIpAddr;
	alias_port = link->insideGlobalPort;

	{
	struct in_addr addr_x;
	 addr_x.s_addr= PKTGET_UINT32_UNALIGNED(&pip->ip_src);
	_rtl8651_l4AliasHandleUdpNbtNS(pip, link, &addr_x, &ud->uh_sport,
								&alias_address, &alias_port);
	}
	return SUCCESS;
}


/*********************************************************************************
* Routine Name :  rtl8651_l4AliasHandleClientUdpNbtIn                                   
* Description NetBios Datagram Handler (IP/UDP):                                                 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
*			before use this function, we should call splnet()
*			afrer this function, we should call splx()
**********************************************************************************/
int32 rtl8651_l4AliasHandleClientUdpNbtIn(struct rtl_pktHdr *pkthdrPtr,
						struct ip *pip, 
					   struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)	
{

	struct udphdr *ud;
	struct in_addr original_address;

	ud = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl & 0xf) << 2));

	original_address.s_addr = link->insideLocalIpAddr;
	ud->uh_dport = link->insideLocalPort;

	return _rtl8651_l4AliasHandleUdpNbt(pip, link, &original_address, ud->uh_dport);
}


/*********************************************************************************
* Routine Name :   rtl8651_l4AliasHandleUdpNbtNSIn                              
* Description :                                                 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
*			before use this function, we should call splnet()
*			afrer this function, we should call splx()
**********************************************************************************/
int32 rtl8651_l4AliasHandleClientUdpNbtNSIn(struct rtl_pktHdr *pkthdrPtr,
						struct ip *pip, 
					   struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link)	
{
	struct udphdr *ud;
	uint16    alias_port;
	struct in_addr alias_address;
	struct in_addr original_address;


	ud = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl & 0xf) << 2));

	alias_address.s_addr = link->insideGlobalIpAddr;
	original_address.s_addr = link->insideLocalIpAddr;
	alias_port = ud->uh_dport;
	ud->uh_dport = link->insideLocalPort;
	
	return _rtl8651_l4AliasHandleUdpNbtNS(pip, link, &alias_address, &alias_port,
									&original_address, &ud->uh_dport);
}



/*********************************************************************************
* Routine Name :  _rtl8651_l4AliasHandleUdpNbt                                   
* Description NetBios Datagram Handler (IP/UDP):                                                 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static int32 _rtl8651_l4AliasHandleUdpNbt(struct ip *pip,	/* IP packet to examine/patch */
						struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s *link,
						struct in_addr *alias_address, uint16 alias_port)
{
	struct udphdr *uh;
	NbtDataHeader *ndh;
	uint8    *p = NULL;
	int8     *pmax;
	int8	     nbtname[MAX_NBT_NAME_LEN]; /* maximum name length is 16; */

	/*
	   Calculate data length of UDP packet 
	 */
	uh = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl & 0xf) << 2));
	pmax = (int8 *) uh + ntohs(uh->uh_ulen);

	ndh = (NbtDataHeader *) ((int8 *) uh + (sizeof(struct udphdr)));
	if ((int8 *) (ndh + 1) > pmax)
		return (-1);

	switch (ndh->type)
	{
		case DGM_DIRECT_UNIQ:		
		case DGM_DIRECT_GROUP:
		case DGM_BROADCAST:
			p = (uint8 *) ndh + 14;			
			p = _rtl8651_l4AliasHandleName(p, pmax,nbtname,MAX_NBT_NAME_LEN);	/* Source Name */
			p = _rtl8651_l4AliasHandleName(p, pmax,nbtname,MAX_NBT_NAME_LEN );	/* Destination Name */
			break;
		case DGM_ERROR:
			p = (uint8 *) ndh + 11;
			break;
		case DGM_QUERY:
		case DGM_POSITIVE_RES:
		case DGM_NEGATIVE_RES:
			p = (uint8 *) ndh + 10;
			p = _rtl8651_l4AliasHandleName(p, pmax,nbtname,MAX_NBT_NAME_LEN);	/* Destination Name */
			break;
	}
	if (p == NULL || (int8 *) p > pmax)
		p = NULL;
	
	/* Doing a IP address and Port number Translation */
	if (uh->uh_sum != 0)
	{
		int32     acc,tmp_val;
		uint16   *sptr;

		/* l3 checksum */
		sptr = (uint16 *) & (ndh->source_ip);
		acc = *sptr++;
		acc += *sptr;
		sptr = (uint16 *) alias_address;
		acc -= *sptr++;
		acc -= *sptr;
		tmp_val = acc;
		_RTL8651_ADJUST_CHECKSUM(acc, pip->ip_sum);

		/* l4 checksum */
		tmp_val += ndh->source_port;
		tmp_val -= alias_port;
		_RTL8651_ADJUST_CHECKSUM(tmp_val, uh->uh_sum);
	}
	
	rtl8651_memcpy(&ndh->source_ip , alias_address,4);	
	ndh->source_port = alias_port;

	return ((p == NULL) ? -1 : 0);
}


/*********************************************************************************
* Routine Name :   _rtl8651_l4AliasHandleUdpNbtNS                                  
* Description :                                                 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
*			before use this function, we should call splnet()
*			afrer this function, we should call splx()
**********************************************************************************/
static int32 _rtl8651_l4AliasHandleUdpNbtNS(struct ip * pip,	/* IP packet to examine/patch */
						  struct rtl8651_tblDrv_naptTcpUdpFlowEntry_s * link,
						  struct in_addr * alias_address,
						  uint16 * alias_port,
						  struct in_addr * original_address,
						  uint16 * original_port)
{
	struct udphdr *uh;
	NbtNSHeader *nsh;
	uint8    *p;
	int8     *pmax;
	NBTArguments nbtarg;

	/*
	   Set up Common Parameter 
	 */
	nbtarg.oldaddr = *alias_address;
	nbtarg.oldport = *alias_port;
	nbtarg.newaddr = *original_address;
	nbtarg.newport = *original_port;
	nbtarg.pip = pip;

	/*
	   Calculate data length of UDP packet 
	 */
	uh = (struct udphdr *) ((int8 *) pip + ((pip->ip_vhl & 0xf) << 2));
	nbtarg.uh_sum = &(uh->uh_sum);
	nsh = (NbtNSHeader *) ((int8 *) uh + (sizeof(struct udphdr)));
	p = (uint8 *) (nsh + 1);
	pmax = (int8 *) uh + ntohs(uh->uh_ulen);

	if ((int8 *) (nsh + 1) > pmax)
		return (-1);

	/* Question Entries */
	if (ntohs(nsh->qdcount) != 0)
	{
		p = _rtl8651_l4AliasHandleQuestion(ntohs(nsh->qdcount),
								(NBTNsQuestion *) p, pmax, &nbtarg);
	}

	/* Answer Resource Records  */
	if (ntohs(nsh->ancount) != 0)
	{
		p = _rtl8651_l4AliasHandleResource(ntohs(nsh->ancount),
								(NBTNsResource *) p, pmax, &nbtarg);
	}

	/* Authority Resource Recodrs */
	if (ntohs(nsh->nscount) != 0)
	{
		p = _rtl8651_l4AliasHandleResource(ntohs(nsh->nscount),
								(NBTNsResource *) p, pmax, &nbtarg);
	}

	/*
	   Additional Resource Recodrs 
	 */
	if (ntohs(nsh->arcount) != 0)
	{
		p = _rtl8651_l4AliasHandleResource(ntohs(nsh->arcount),
								(NBTNsResource *) p, pmax, &nbtarg);
	}

	return ((p == NULL) ? -1 : 0);
}


/*********************************************************************************
* Routine Name :  _rtl8651_l4AliasHandleName                                   
* Description : Handling Name field                                                 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static uint8 *_rtl8651_l4AliasHandleName(uint8 * p, int8 * pmax, int8 *nbtname, int32 nbtnamelen)
{

	uint8    *s;
	uint8     c;
	int32     compress;
	int32     ret_len = 0;
	int32     i;

	/* Following length field */

	if (p == NULL || (int8 *) p >= pmax)
		return (NULL);

	if (*p & 0xc0)
	{
		p = p + 2;
		if ((int8 *) p > pmax)
			return (NULL);
		return ((uint8 *) p);
	}
	while ((*p & 0x3f) != 0x00)
	{
		s = p + 1;
		if (*p == 0x20)
			compress = 1;
		else
			compress = 0;

		/*
		   Get next length field 
		 */
		p = (uint8 *) (p + (*p & 0x3f) + 1);
		if ((int8 *) p > pmax)
		{
			p = NULL;
			break;
		}
		while (s < p)
		{
			if (compress == 1)
			{
				////c = (uint8) (((((*s & 0x0f) << 4) | (*(s + 1) & 0x0f)) -
				////			  0x11));

				c = ( (( (int)(*s) - 'A' ) << 4) +  ( (int)(*(s+1)) - 'A' ) );
								
				if(ret_len < (nbtnamelen-1) )
					nbtname[ret_len] = c;
				else
					nbtname[nbtnamelen-1] = '\0';
				
				ret_len++;
				s += 2;

				/* 
				The RFCs do not say so, but the last byte of the NetBIOS name is reserved.
				The practice probably goes back to the early days and IBM. 
				The 16th byte of a NetBIOS name is used to designate the purpose 
				of the name. This byte is known as the "suffix" (or sometimes the "type byte"), 
				and it contains a value which indicates the type of service that registered the name. 
				Some example suffix values include: 

				0x00 ==  Workstation Service (aka. Machine Name or Client Service) 
				0x03 ==  Messenger Service 
				0x20 ==  File Server Service 
				0x1B ==  Domain Master Browser 
				*/
				nbtname[15] = '\0';
								
				
			}
			else
			{
				if(ret_len < (nbtnamelen-1) )
					nbtname[ret_len] = (int8)(*s);
				else
					nbtname[nbtnamelen-1] = '\0';		
				
				ret_len++;				
				s++;

				/* 
				The RFCs do not say so, but the last byte of the NetBIOS name is reserved.
				The practice probably goes back to the early days and IBM. 
				The 16th byte of a NetBIOS name is used to designate the purpose 
				of the name. This byte is known as the "suffix" (or sometimes the "type byte"), 
				and it contains a value which indicates the type of service that registered the name. 
				Some example suffix values include: 

				0x00 ==  Workstation Service (aka. Machine Name or Client Service) 
				0x03 ==  Messenger Service 
				0x20 ==  File Server Service 
				0x1B ==  Domain Master Browser 
				*/
				nbtname[15] = '\0';
				
			}
		}
	}

	/*
	   Set up to out of Name field 
	 */
	if (p == NULL || (int8 *) p >= pmax)
		p = NULL;
	else
		p++;

	/* padding '\0' */
	if(ret_len < (nbtnamelen-1) )
		nbtname[ret_len] = '\0';
	else
		nbtname[nbtnamelen-1] = '\0';		

	/* Trim off trailing spaces, if any. */
	  for( i = (nbtnamelen-1); (i >= 0) && (' ' == nbtname[i]); i-- )
	    nbtname[i] = '\0';
	
	
	return ((uint8 *) p);
}


/*********************************************************************************
* Routine Name : _rtl8651_l4AliasHandleQuestion                                    
* Description : Question Section                                                
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
*			before use this function, we should call splnet()
*			afrer this function, we should call splx()
**********************************************************************************/
static uint8 *_rtl8651_l4AliasHandleQuestion(uint16 count,
								  NBTNsQuestion * q, int8 * pmax,
								  NBTArguments * nbtarg)
{

	int8 nbtname[MAX_NBT_NAME_LEN];
	
	while (count != 0)
	{
		/*
		   Name Filed 
		 */
		q = (NBTNsQuestion *) _rtl8651_l4AliasHandleName((uint8 *) q, pmax,nbtname,MAX_NBT_NAME_LEN);

		if (q == NULL || (int8 *) (q + 1) > pmax)
		{
			q = NULL;
			break;
		}

		/*
		   Type and Class filed 
		 */
		switch (ntohs(q->type))
		{
			case QS_TYPE_NB:
			case QS_TYPE_NBSTAT:
				q = q + 1;
				break;
			default:
				break;
		}
		count--;
	}

	/*
	   Set up to out of Question Section 
	 */
	return ((uint8 *) q);
}


/*********************************************************************************
* Routine Name :   AliasHandleResourceNB
* Description :   Resource Record                                             
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
*			before use this function, we should call splnet()
*			afrer this function, we should call splx()
**********************************************************************************/
static uint8 *AliasHandleResourceNB(NBTNsResource * q, int8 * pmax,
									NBTArguments * nbtarg)
{
	NBTNsRNB *nb;
	uint16    bcount;

	if (q == NULL || (int8 *) (q + 1) > pmax)
		return (NULL);
	/*
	   Check out a length 
	 */
	bcount = ntohs(q->rdlen);

	/*
	   Forward to Resource NB position 
	 */
	nb = (NBTNsRNB *) ((uint8 *) q + SizeOfNsResource);

	/*
	   Processing all in_addr array 
	 */
	while (nb != NULL && bcount != 0)
	{
		if ((int8 *) (nb + 1) > pmax)
		{
			nb = NULL;
			break;
		}
		if (!memcmp(&nbtarg->oldaddr, &nb->addr, sizeof(struct in_addr)))
		{
			if (*nbtarg->uh_sum != 0)
			{
				int32     acc,tmp_val;
				uint16   *sptr;

				/* l4 checksum */
				sptr = (uint16 *) & (nb->addr);
				acc = *sptr++;
				acc += *sptr;
				sptr = (uint16 *) & (nbtarg->newaddr);
				acc -= *sptr++;
				acc -= *sptr;
				tmp_val = acc;
				_RTL8651_ADJUST_CHECKSUM(acc, nbtarg->uh_sum);			

				/* l3 checksum */
				_RTL8651_ADJUST_CHECKSUM(tmp_val, nbtarg->pip->ip_sum);	
			}
			
			rtl8651_memcpy(&nb->addr ,&nbtarg->newaddr,4);
		}
		nb = (NBTNsRNB *) ((uint8 *) nb + SizeOfNsRNB);
		bcount -= SizeOfNsRNB;
	}
	if (nb == NULL || (int8 *) (nb + 1) > pmax)
	{
		nb = NULL;
	}

	return ((uint8 *) nb);
}


/*********************************************************************************
* Routine Name :  AliasHandleResourceA                                   
* Description :                                                 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static uint8 *AliasHandleResourceA(NBTNsResource * q, int8 * pmax,
								   NBTArguments * nbtarg)
{
	NBTNsResourceA *a;
	uint16    bcount;

	if (q == NULL || (int8 *) (q + 1) > pmax)
		return (NULL);

	/*
	   Forward to Resource A position 
	 */
	a = (NBTNsResourceA *) ((uint8 *) q + sizeof(NBTNsResource));

	/*
	   Check out of length 
	 */
	bcount = ntohs(q->rdlen);

	/*
	   Processing all in_addr array 
	 */
	while (bcount != 0)
	{
		if (a == NULL || (int8 *) (a + 1) > pmax)
			return (NULL);

		if (!memcmp(&nbtarg->oldaddr, &a->addr, sizeof(struct in_addr)))
		{
			if (*nbtarg->uh_sum != 0)
			{
				int32     acc,tmp_val;
				uint16   *sptr;

				/* l4 checksum */
				sptr = (uint16 *) & (a->addr);	/* Old */
				acc = *sptr++;
				acc += *sptr;
				sptr = (uint16 *) & nbtarg->newaddr;	/* New */
				acc -= *sptr++;
				acc -= *sptr;
				tmp_val = acc;
				_RTL8651_ADJUST_CHECKSUM(acc, nbtarg->uh_sum);	
				
				/* l3 checksum */
				_RTL8651_ADJUST_CHECKSUM(tmp_val, nbtarg->pip->ip_sum);					
				
			}
			rtl8651_memcpy(&a->addr, &nbtarg->newaddr,4);
		}
		a++;
		 /*XXXX*/ bcount -= SizeOfResourceA;
	}
	if (a == NULL || (int8 *) (a + 1) > pmax)
		a = NULL;
	return ((uint8 *) a);
}


/*********************************************************************************
* Routine Name : AliasHandleResourceNULL                                    
* Description :                                                 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
*			before use this function, we should call splnet()
*			afrer this function, we should call splx()
**********************************************************************************/
static uint8 *AliasHandleResourceNULL(NBTNsResource * q,
									  int8 * pmax, NBTArguments * nbtarg)
{
	NBTNsResourceNULL *n;
	uint16    bcount;

	if (q == NULL || (int8 *) (q + 1) > pmax)
		return (NULL);

	/*
	   Forward to Resource NULL position 
	 */
	n = (NBTNsResourceNULL *) ((uint8 *) q + sizeof(NBTNsResource));

	/*
	   Check out of length 
	 */
	bcount = ntohs(q->rdlen);

	/*
	   Processing all in_addr array 
	 */
	while (bcount != 0)
	{
		if ((int8 *) (n + 1) > pmax)
		{
			n = NULL;
			break;
		}
		n++;
		bcount -= sizeof(NBTNsResourceNULL);
	}
	if ((int8 *) (n + 1) > pmax)
		n = NULL;

	return ((uint8 *) n);
}


/*********************************************************************************
* Routine Name :   AliasHandleResourceNS                                  
* Description :                                                 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
*			before use this function, we should call splnet()
*			afrer this function, we should call splx()
**********************************************************************************/
static uint8 *AliasHandleResourceNS(NBTNsResource * q, int8 * pmax,
									NBTArguments * nbtarg)
{
	NBTNsResourceNULL *n;
	uint16    bcount;
	int8 nbtname[MAX_NBT_NAME_LEN];

	if (q == NULL || (int8 *) (q + 1) > pmax)
		return (NULL);

	/*
	   Forward to Resource NULL position 
	 */
	n = (NBTNsResourceNULL *) ((uint8 *) q + sizeof(NBTNsResource));

	/*
	   Check out of length 
	 */
	bcount = ntohs(q->rdlen);

	/*
	   Resource Record Name Filed 
	 */
	q = (NBTNsResource *) _rtl8651_l4AliasHandleName((uint8 *) n, pmax,nbtname,MAX_NBT_NAME_LEN);	/* XXX */

	if (q == NULL || (int8 *) ((uint8 *) n + bcount) > pmax)
		return (NULL);
	else
		return ((uint8 *) n + bcount);
}


/*********************************************************************************
* Routine Name :  AliasHandleResourceNBSTAT                                   
* Description :                                                 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
*			before use this function, we should call splnet()
*			afrer this function, we should call splx()
**********************************************************************************/
static uint8 *AliasHandleResourceNBSTAT(NBTNsResource * q,
										int8 * pmax, NBTArguments * nbtarg)
{
	NBTNsResourceNBSTAT *n;
	uint16    bcount;

	if (q == NULL || (int8 *) (q + 1) > pmax)
		return (NULL);

	/*
	   Forward to Resource NBSTAT position 
	 */
	n = (NBTNsResourceNBSTAT *) ((uint8 *) q + sizeof(NBTNsResource));

	/*
	   Check out of length 
	 */
	bcount = ntohs(q->rdlen);

	if (q == NULL || (int8 *) ((uint8 *) n + bcount) > pmax)
		return (NULL);
	else
		return ((uint8 *) n + bcount);
}


/*********************************************************************************
* Routine Name : _rtl8651_l4AliasHandleResource                                     
* Description :                                                 
* Input :                                                               
* Output :                                                             
* Return :                                                             
* Note :   
*        ThreadSafe: n
**********************************************************************************/
static uint8 *_rtl8651_l4AliasHandleResource(uint16 count,
								  NBTNsResource * q, int8 * pmax,
								  NBTArguments * nbtarg)
{
	int8 nbtname[MAX_NBT_NAME_LEN];
	while (count != 0)
	{
		/* Resource Record Name Filed */
		q = (NBTNsResource *) _rtl8651_l4AliasHandleName((uint8 *) q, pmax,nbtname,MAX_NBT_NAME_LEN);

		if (q == NULL || (int8 *) (q + 1) > pmax)
			break;

		/* Type and Class filed */
		switch (ntohs(q->type))
		{
			case RR_TYPE_NB:
				q = (NBTNsResource *) AliasHandleResourceNB(q, pmax, nbtarg);
				break;
			case RR_TYPE_A:
				q = (NBTNsResource *) AliasHandleResourceA(q, pmax, nbtarg);
				break;
			case RR_TYPE_NS:
				q = (NBTNsResource *) AliasHandleResourceNS(q, pmax, nbtarg);
				break;
			case RR_TYPE_NULL:
				q = (NBTNsResource *) AliasHandleResourceNULL(q, pmax,
															  nbtarg);
				break;
			case RR_TYPE_NBSTAT:
				q = (NBTNsResource *) AliasHandleResourceNBSTAT(q, pmax,
																nbtarg);
				break;
			default:
				break;
		}
		count--;
	}

	return ((uint8 *) q);
}


