
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <linux/videodev.h>
#include "pwc-ioctl.h"

static int device_fd;
static struct video_capability vidcap;
static struct video_window vidwin;
static struct video_picture vidpic;
static struct video_clip vidclips[32];


int open_camera(const char *devicename)
{
	
    device_fd = open(devicename, O_RDWR);
    if(device_fd <= 0)
    {
	printf("Device %s couldn't be opened\n", devicename);
	return 0;
    }
    return 1;
}

void close_camera(void)
{
    close(device_fd);
}

void get_camera_info(void)
{
    ioctl(device_fd, VIDIOCGCAP, &vidcap);
    ioctl(device_fd, VIDIOCGWIN, &vidwin);
    ioctl(device_fd, VIDIOCGPICT, &vidpic);
    
    vidwin.clips = vidclips;
    vidwin.clipcount = 0;
}

void print_camera_info(void)
{

    printf("    *** Camera Info ***\n");
    printf("Name:           %s\n", vidcap.name);
    printf("Type:           %d\n", vidcap.type);
    printf("Minimum Width:  %d\n", vidcap.minwidth);
    printf("Maximum Width:  %d\n", vidcap.maxwidth);
    printf("Minimum Height: %d\n", vidcap.minheight);
    printf("Maximum Height: %d\n", vidcap.maxheight);
    printf("X:              %d\n", vidwin.x);
    printf("Y:              %d\n", vidwin.y);
    printf("Width:          %d\n", vidwin.width);
    printf("Height:         %d\n", vidwin.height);
    printf("Depth:          %d\n", vidpic.depth);
 
    if(vidcap.type & VID_TYPE_MONOCHROME)
	printf("Color           false\n");
    else
	printf("Color           true\n");	
}

#if 0     
   

void read_test(int quiet)
{
	int server_sockfd, client_sockfd;	
	struct sockaddr_in server_address;
	struct sockaddr_in client_address;
	//unsigned char *buffer = malloc(vidcap.maxwidth * vidcap.maxheight * 3);
    int len = 0;
	
	server_sockfd = socket(PF_INET, SOCK_STREAM,0);
	bzero(&server_address,sizeof(server_address));	
	server_address.sin_family= PF_INET;
	server_address.sin_addr.s_addr = htonl(INADDR_ANY);
	server_address.sin_port = htons(8888);
	
	bind(server_sockfd,(struct sockaddr *)&server_address, sizeof(server_address));	
		
	listen(server_sockfd,5);
	while(1)
	{
		
		int client_len;
		char command;
		
		struct video_mbuf vidbuf;
    	struct video_mmap vm;
    	int numframe = 0;
    	unsigned char *buffer;
  	
    			
		printf("webcam server waiting...\n");		
		client_len=sizeof(client_address);		
		client_sockfd = accept(server_sockfd,(struct sockaddr *)&client_address, &client_len);
		
    	ioctl(device_fd, VIDIOCGMBUF, &vidbuf);
    	buffer = mmap(NULL, vidbuf.size, PROT_READ, MAP_SHARED, device_fd, 0);

		//printf("vidbuf.size=%d\n",vidbuf.size);
		vm.format = VIDEO_PALETTE_RGB24;			
    	vm.frame  = 0;
    	vm.width  = 352;
    	//vm.width  = 176;
    	vm.height = 292;
    	//vm.height =146;
    	ioctl(device_fd, VIDIOCMCAPTURE, &vm);
		
		while(1)
		{
			//printf("read page\n");
			read(client_sockfd,&command,1);
			if(command==1)
			{		
				//len = read(device_fd, buffer, vidcap.maxwidth * vidcap.maxheight * 3);		
    			ioctl(device_fd, VIDIOCSYNC, &numframe);						
				
				write(client_sockfd, buffer, vidcap.maxwidth * vidcap.maxheight * 3);		
			}
			else
			{
				close(client_sockfd);
				break;
			}
		}
		
	}
	close(server_sockfd);

}
#endif

#if 1


void mmap_test(int quiet)
{
    struct video_mbuf vidbuf;
    struct video_mmap vm;
    unsigned long int numframe = 0;
    int r;
    unsigned char *buffer;
    int filesize;
    int i;
       	FILE *f;
    
    r=ioctl(device_fd, VIDIOCGMBUF, &vidbuf);
    if(r<0) printf("ioctl error\n");
    buffer = mmap(NULL, vidbuf.size, PROT_READ, MAP_SHARED, device_fd, 0);
 

//    vm.format = VIDEO_PALETTE_RGB24;
    vm.format = VIDEO_PALETTE_YUV420P;
    vm.frame  = 0;
  
    vm.width  = 640;
    vm.height = 480;
    

    r=ioctl(device_fd, VIDIOCMCAPTURE, &vm);
    if(r!=0) printf("ioctl error2\n");

    r=ioctl(device_fd, VIDIOCSYNC, &numframe);
    if(r!=0) printf("ioctl error3\n");
    
    //check image size, high performance algorithm
    {
    	unsigned long int now,high=40960,low=0;    	
    	while(1)
    	{
    		now=high;
    		if((buffer[now]!=0x80)&&(buffer[now+1]!=0x80)) break;
    		while((buffer[now]==0x80)&&(buffer[now+1]==0x80))
    		{
    			high=now;
    			now=((now-low)>>1)+low;    		
    		}
    		low=now;
    		
    		if((low+1)>=high) break;    		
    	}    	
    	filesize=high;
    	printf("file size=%x\n",high);    	
    }


    
 
	//hexdump_data(buffer, filesize);
	f=fopen("/var/tp6800.jpg","wb");
	fwrite(buffer,filesize,1,f);
	fclose(f);	
   
  
}
#endif

int main(int argc, char *argv[])
{
	FILE *f;
	f=fopen("/var/run/quickcam.pid","w");
	fprintf(f,"%d",getpid());
	fclose(f);

  	if(open_camera("/dev/video0") == 1)
	{
		get_camera_info();
		print_camera_info();
		//read_test(0);
		mmap_test(0);
		close_camera();
	}
	
	unlink("/var/run/quickcam.pid");
	return 0;  
}

